import numpy as np

from gathers import Gathers

NUM = 1000
CLUSTER = 10
DIM = 16
RABITQ_MATCH_RATE = 0.99


def test_rabitq():
    gathers = Gathers(verbose=True)
    rng = np.random.default_rng()
    arr = rng.random((NUM, DIM), dtype=np.float32)
    c = gathers.fit(arr, CLUSTER)
    assert c.shape == (CLUSTER, DIM), c.shape

    # test `assign`
    for vec in arr:
        distances = np.linalg.norm(c - vec, axis=1)
        assert np.argmin(distances) == gathers.assign(vec, c)

    # test `batch_assign`
    labels = gathers.batch_assign(arr, c)
    assert len(labels) == len(arr)
    expect = [np.argmin(np.linalg.norm(c - vec, axis=1)) for vec in arr]
    match_rate = np.sum(np.array(expect) == np.array(labels)) / NUM
    assert match_rate > RABITQ_MATCH_RATE, match_rate



from pathlib import Path
from struct import pack, unpack

import numpy as np


def read_vec(filepath: str, vec_type: np.dtype = np.float32):
    """Read vectors from a file. Support `fvecs`, `ivecs` and `bvecs` format.
    Args:
        filepath: The path of the file.
        vec_type: The type of the vectors.
    """
    size = np.dtype(vec_type).itemsize
    with open(filepath, "rb") as f:
        vecs = []
        while True:
            try:
                buf = f.read(4)
                if len(buf) == 0:
                    break
                dim = unpack("<i", buf)[0]
                vecs.append(np.frombuffer(f.read(dim * size), dtype=vec_type))
            except Exception as err:
                print(err)
                break
    return np.array(vecs)


def write_vec(filepath: str, vecs: np.ndarray, vec_type: np.dtype = np.float32):
    """Write vectors to a file. Support `fvecs`, `ivecs` and `bvecs` format."""
    with open(filepath, "wb") as f:
        for vec in vecs:
            f.write(pack("<i", len(vec)))
            f.write(vec.tobytes())


def generate_random(dim: int, num: int, query: int, topk: int, filepath: str):
    rng = np.random.default_rng()
    x = rng.integers(0, 256, (num, dim), dtype=np.uint8).astype(np.float32)
    mean = x.mean(axis=0)
    q = rng.integers(0, 256, (query, dim), dtype=np.uint8).astype(np.float32)
    # labels = [[np.argmin(np.linalg.norm(x - vec, axis=1))] for vec in q]
    labels = [
        np.argpartition(np.linalg.norm(x - vec, axis=1), topk)[:topk] for vec in q
    ]

    path = Path(filepath)
    path.mkdir(parents=True, exist_ok=True)
    write_vec(path / "base.fvecs", x)
    write_vec(path / "centroids.fvecs", np.array([mean]))
    write_vec(path / "query.fvecs", q)
    write_vec(path / "groundtruth.ivecs", np.array(labels, dtype=np.int32))


if __name__ == "__main__":
    generate_random(100, 1000, 10, 1, "dataset-random")



from argparse import ArgumentParser
from time import perf_counter
from struct import unpack, pack

from faiss import Kmeans
from sklearn.cluster import MiniBatchKMeans
import numpy as np


def build_arg_parser():
    parser = ArgumentParser()
    parser.add_argument("--n_clusters", "-n", type=int, default=4096)
    parser.add_argument("--max_iter", "-m", type=int, default=25)
    parser.add_argument("--verbose", "-v", action="store_true")
    parser.add_argument("--input", "-i", type=str, required=True)
    parser.add_argument("--output", "-o", type=str, required=True)
    parser.add_argument(
        "--library", "-l", type=str, default="faiss", choices=["faiss", "sklearn"]
    )
    return parser


def read_vec(filepath: str, vec_type: np.dtype = np.float32):
    """Read vectors from a file. Support `fvecs`, `ivecs` and `bvecs` format.
    Args:
        filepath: The path of the file.
        vec_type: The type of the vectors.
    """
    size = np.dtype(vec_type).itemsize
    with open(filepath, "rb") as f:
        vecs = []
        while True:
            try:
                buf = f.read(4)
                if len(buf) == 0:
                    break
                dim = unpack("<i", buf)[0]
                vecs.append(np.frombuffer(f.read(dim * size), dtype=vec_type))
            except Exception as err:
                print(err)
                break
    return np.array(vecs)


def write_vec(filepath: str, vecs: np.ndarray, vec_type: np.dtype = np.float32):
    """Write vectors to a file. Support `fvecs`, `ivecs` and `bvecs` format."""
    with open(filepath, "wb") as f:
        for vec in vecs.astype(vec_type):
            f.write(pack("<i", len(vec)))
            f.write(vec.tobytes())


def faiss_cluster(args):
    vecs = read_vec(args.input)
    dim = vecs.shape[1]
    kmeans = Kmeans(dim, args.n_clusters, niter=args.max_iter, verbose=args.verbose)
    t_start = perf_counter()
    kmeans.train(vecs)
    print(f"faiss k-means training time: {perf_counter() - t_start:.6f}s")
    write_vec(args.output, kmeans.centroids)


def sklearn_cluster(args):
    vecs = read_vec(args.input)
    kmeans = MiniBatchKMeans(
        n_clusters=args.n_clusters, max_iter=args.max_iter, verbose=args.verbose
    )
    t_start = perf_counter()
    kmeans.fit(vecs)
    print(
        f"scikit-learn k-means training + assign time: {perf_counter() - t_start:.6f}s"
    )
    write_vec(args.output, kmeans.cluster_centers_)


if __name__ == "__main__":
    args = build_arg_parser().parse_args()
    print(args)
    if args.library == "faiss":
        faiss_cluster(args)
    else:
        sklearn_cluster(args)



from __future__ import annotations

from os import environ

import numpy as np

from .gatherspy import assign, batch_assign, kmeans_fit

__all__ = ["Gathers"]

MATRIX_SHAPE = 2


class Gathers:
    def __init__(self, verbose: bool = False):
        if verbose:
            environ["GATHERS_LOG"] = "debug"

    def assign(self, vec: np.ndarray, centroids: np.ndarray) -> int:
        """
        Assign the vector to the nearest centroid.

        This method is slower than :py:meth:`~Gathers.batch_assign`, but it's
        100% accurate.
        """
        assert (
            len(vec.shape) == 1
            and len(centroids.shape) == MATRIX_SHAPE
            and vec.shape[0] == centroids.shape[1]
        )
        return assign(vec, centroids)

    def batch_assign(self, vecs: np.ndarray, centroids: np.ndarray) -> list[int]:
        """
        Assign each vector to the nearest centroid with RaBitQ.

        This method does not guarantee it's 100% accurate, but it's usually faster,
        especially when the dim is large.

        Returns:
            list[int]: The list of the assigned labels.
        """
        assert (
            len(vecs.shape) == MATRIX_SHAPE
            and len(centroids.shape) == MATRIX_SHAPE
            and vecs.shape[1] == centroids.shape[1]
        )
        return batch_assign(vecs, centroids)

    def fit(self, vecs: np.ndarray, n_cluster: int, max_iter: int = 25) -> np.ndarray:
        """
        Cluster the vectors into `n_cluster` clusters with max iteration `max_iter`.

        Returns:
            np.ndarray: The centroids of the clusters.
        """
        assert len(vecs.shape) == MATRIX_SHAPE
        return kmeans_fit(vecs, n_cluster, max_iter)
